/// This routine integrates the system in time. An array of snapshots taken at each time specified
/// by the times vector is returned. 

#include "swift.h"
#include "assert.h"
#include "systemic.h"
#include "utils.h"
#include "integration.h"
#include "stdint.h"
#include "unistd.h"

#define NTPMAX 1001
#define NPLMAX 51
#define NSTATP 3
#define NSTAT NSTATP + NPLMAX - 1

#define NSTATR NSTAT

typedef void(*swift_integrator)(int*, double*, int*, int*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        int*, double*, double*);

// Bulirsch-Stoer
extern void bs_step_(int*, double*, int*, int*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        int*, double*, double*);

// 
extern void step_kdk_(int*, double*, int*, int*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        int*, double*, double*);

extern void rmvs3_step_(int*, double*, int*, int*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        double*, double*, double*, double*, double*, double*,
        int*, double*, double*);

swift_integrator sis[3] = { bs_step_, step_kdk_, rmvs3_step_};

ok_system** ok_integrate_swift(ok_system* initial, const gsl_vector* times, const ok_integrator_options* options,
    int si, ok_system** bag) {
    
    const double startTime = initial->epoch;
    int NDIMS = initial->nplanets + 1;
    
    // Allocate the return array of snapshots
    const int SAMPLES = times->size;
    if (bag == NULL) {
        bag = (ok_system**) calloc(SAMPLES, sizeof(ok_system*));
        #pragma omp parallel for
        for (int i = 0; i < SAMPLES; i++) {
            bag[i] = ok_copy_system(initial);
        }
    } else {
        #pragma omp parallel for
        for (int i = 0; i < SAMPLES; i++) {
            bag[i]->epoch = initial->epoch;
            bag[i]->flag = initial->flag;
            gsl_matrix_memcpy(bag[i]->elements, initial->elements);
        }
    }

    gsl_matrix* prevOrbits = initial->orbits;
    double prevTime = startTime;
    
    
    double rstat[NSTATR][NTPMAX];
    int istat[NSTAT][NTPMAX];
    int ostdin = 0;
    
    if (si == 0) {
        // The Bulirsch-Stoer integrator asks for eps directly from the
        // standard input. We're going to manually feed it :-/
        // Unfortunately, this is not portable ANSI C: I'm making use
        // of dup() and dup2() to redirect and restore stdin.
        double eps = options->rel_acc;
        fflush(stdin);
        ostdin = dup(fileno(stdin));
        char nam[L_tmpnam];
        tmpnam(nam);
        FILE* tmp = fopen(nam, "w");
        fprintf(tmp, "%e\n", eps);
        fclose(tmp);
        
        freopen(nam, "r", stdin);
    }
    
    swift_integrator sif = sis[si];
    
    double* x, *y, *z, *vx, *vy, *vz, *mass;
    if (options != NULL && options->buffer != NULL) {
        x = options->buffer[1];
        y = options->buffer[2];
        z = options->buffer[3];
        vx = options->buffer[4];
        vy = options->buffer[5];
        vz = options->buffer[6];
        mass = options->buffer[0];
    } else {
        x = (double*) malloc(sizeof(double) * NDIMS);
        y = (double*) malloc(sizeof(double) * NDIMS);
        z = (double*) malloc(sizeof(double) * NDIMS);
        vx = (double*) malloc(sizeof(double) * NDIMS);
        vy = (double*) malloc(sizeof(double) * NDIMS);
        vz = (double*) malloc(sizeof(double) * NDIMS);
        mass = (double*) malloc(sizeof(double) * NDIMS);
    }
    
    gsl_matrix* xyz = initial->xyz;
    #pragma omp parallel for
    for (int i = 0; i < NDIMS; i++) {
        mass[i] = MGET(xyz, i, 0);
        x[i] = MGET(xyz, i, 1) - MGET(xyz, 0, 1);
        y[i] = MGET(xyz, i, 2) - MGET(xyz, 0, 2);
        z[i] = MGET(xyz, i, 3) - MGET(xyz, 0, 3);
        vx[i] = MGET(xyz, i, 4) - MGET(xyz, 0, 4);
        vy[i] = MGET(xyz, i, 5) - MGET(xyz, 0, 5);
        vz[i] = MGET(xyz, i, 6) - MGET(xyz, 0, 6);
    }
    
    double dum[1];
    int dum0 = 0;
    double dum0d = 0.;
    
    
    int step = 0;
    // Loop through the times vector
    for (int i = 0; i < SAMPLES; i++) {
        double time = times->data[i];
        double dt = time - prevTime;
        // Integrate between prevTime and time
        if (fabs(time - prevTime) > 1e-10) {
                // for now, no test particles/J2/J4
                sif(&step, &prevTime, &NDIMS, &dum0, mass, &dum0d, &dum0d, 
                        x, y, z, vx, vy, vz, 
                        dum, dum, dum, dum, dum, dum,
                        (int*) istat, (double*) rstat, &dt);
               
                if (si == 0) {
                    dup2(ostdin, fileno(stdin));
                    close(ostdin);
                }
        }
        

        // Set up the return vector
        bag[i]->time = time;
        
        // Allocates space to save the xyz coordinates in the snapshot, and 
        // copies the current coordinates.
        gsl_matrix* nxyz = gsl_matrix_alloc(NDIMS, 7);
        
        #pragma omp parallel for
        for (int j = 0; j < NDIMS; j++) {
            MSET(nxyz, j, 0, mass[j]);
            MSET(nxyz, j, 1, x[j]);
            MSET(nxyz, j, 2, y[j]);
            MSET(nxyz, j, 3, z[j]);
            MSET(nxyz, j, 4, vx[j]);
            MSET(nxyz, j, 5, vy[j]);
            MSET(nxyz, j, 6, vz[j]);
        }
        bag[i]->xyz = nxyz;
        
        // Converts the current cartesian coordinates back to orbital elements
        bag[i]->orbits = gsl_matrix_alloc(NDIMS, ELEMENTS_SIZE);
        gsl_matrix_memcpy(bag[i]->orbits, prevOrbits);
        ok_cart2el(bag[i], bag[i]->orbits, true);
        prevOrbits = bag[i]->orbits;
        
        prevTime = time;
    }
    
    if (options == NULL || options->buffer == NULL) {
        free(x);
        free(y);
        free(z);
        free(vx);
        free(vy);
        free(vz);
        free(mass);   
    }
    return bag;
}

